// 2019-08-02 ffgl@subpixels.com
// - Make inline shims for Compile(), FindUniform(), Set() and Bind() with const string& arguments
// - Add variants of Set() with uniformLocation arguments
// - Replace CompileVertexShader(), CompileGeometryShader(), CompileFragmentShader() by generic CompileShader()

#pragma once
#include "../ffgl/FFGL.h"//For OpenGL
#include <vector>
#include <string>
#include <map>
#include <string>

namespace ffglex
{
/**
 * The FFGLShader is a helper class you can use to compile glsl shader strings
 * and link them together into a shader program. You can then get the program id
 * and use it to bind the resulting shader to use it for the next rendering operation.
 */
class FFGLShader final
{
public:
	FFGLShader(); //The default constructor just initializes this object to represent no shader.
	~FFGLShader();//Destructing this object just checks and lets you know if you've properly released the GL resources.

	void AddTransformFeedbackVarying( const std::string& varyingName );                              //Add a varying that will be captured using a transform feedback.
	bool Compile( const char* vertexShader, const char* fragmentShader );                            //Compiles and links two shaders into a shader program.
	bool Compile( const char* vertexShader, const char* geometryShader, const char* fragmentShader );//Compiles and links a vertex/geometry/fragment shader into a single shader program.
	void FreeGLResources();                                                                          //Frees any GL resources that this shader is holding.

	bool IsReady() const;                       //Whether or not previous complilation succeeded and this shader is ready to be used for rendering.
	void Use() const;                           //Call glUseProgram with the current programID
	GLuint GetGLID() const;                     //Gets the OpenGL ID that represents the linked shader program.
	GLint FindUniform( const char* name ) const;//Finds the uniform location that needs to be used to addess a uniform with a certain name.

	void Set( const char* name, float value );
	void Set( const char* name, float v1, float v2 );
	void Set( const char* name, float v1, float v2, float v3 );
	void Set( const char* name, float v1, float v2, float v3, float v4 );
	void Set( const char* name, int value );

	void Set( GLint uniformLocation, float value );
	void Set( GLint uniformLocation, float v1, float v2 );
	void Set( GLint uniformLocation, float v1, float v2, float v3 );
	void Set( GLint uniformLocation, float v1, float v2, float v3, float v4 );
	void Set( GLint uniformLocation, int value );

	void Bind( const char* name, int texture, const FFGLTextureStruct& fbo );

	inline bool Compile(const std::string& vertexShader, const std::string& fragmentShader)
	{
		return Compile( vertexShader.c_str(), fragmentShader.c_str() );
	}
	inline bool Compile( const std::string& vertexShader, const std::string& geometryShader, const std::string& fragmentShader )
	{
		return Compile( vertexShader.c_str(), geometryShader.c_str(), fragmentShader.c_str() );
	}

	inline GLint FindUniform( const std::string& name ) const
	{
		return FindUniform( name.c_str() );
	}

	inline void Set( const std::string& name, float value )
	{
		Set( name.c_str(), value );
	}
	inline void Set( const std::string& name, float v1, float v2 )
	{
		Set( name.c_str(), v1, v2 );
	}
	inline void Set( const std::string& name, float v1, float v2, float v3 )
	{
		Set( name.c_str(), v1, v2, v3 );
	}
	inline void Set( const std::string& name, float v1, float v2, float v3, float v4 )
	{
		Set( name.c_str(), v1, v2, v3, v4 );
	}
	inline void Set( const std::string& name, int value )
	{
		Set( name.c_str(), value );
	}

	inline void Bind(const std::string& name, int texture, const FFGLTextureStruct& fbo)
	{
		Bind( name.c_str(), texture, fbo );
	}

	FFGLShader& operator=( const FFGLShader& ) = delete;
	FFGLShader& operator=( FFGLShader&& ) = delete;

private:
	bool CompileShader( const GLuint shaderID, const char* shaderCode, const char* shaderName );//Internal utility function that tries to compile a shader. Doesn't clean up on failure.
	bool LinkProgram();                                                                         //Internal utility function that tries to link the two previously compiled vertex and fragment shaders into a program. Doesn't clean up on failure.

	GLuint vertexShaderID;                               //!< The ID OpenGL gave our vertex shader. 0 for invalid.
	GLuint geometryShaderID;                             //!< The ID OpenGL gave our geometry shader. 0 if not present or invalid.
	GLuint fragmentShaderID;                             //!< The ID OpenGL gave our fragment shader. 0 for invalid.
	GLuint programID;                                    //!< The ID OpenGL gave our shader program. Bind this to use this shader. 0 for invalid.
	std::vector< std::string > transformFeedbackVaryings;//!< The varyings that will be captured using a transform feedback. Ordered in the order of capturing.
};

}//End namespace ffglex