#include "Plugin.h"
#include <assert.h>

#define LOGSHADERNOTES

using namespace ffglex;

Plugin::Plugin()
{
}

Plugin::~Plugin()
{
	for( Param* param : params )
	{
		if( param != nullptr )
			delete param;
	}
}

FFResult Plugin::InitGL( const FFGLViewportStruct* viewPort )
{
	std::string fragmentShaderCode = createFragmentShader( fragmentShaderBase );
	if( !shader.Compile( vertexShaderCode, fragmentShaderCode ) )
	{
		DeInitGL();
		return FF_FAIL;
	}

	shader.Use();
	setParamaterUniformLocations();

	if( !quad.Initialise() )
	{
		DeInitGL();
		return FF_FAIL;
	}
	if( init() == FF_FAIL )
	{
		DeInitGL();
		return FF_FAIL;
	}
	resetOpenGLState();
	return CFreeFrameGLPlugin::InitGL( viewPort );
}

FFResult Plugin::ProcessOpenGL( ProcessOpenGLStruct* inputTextures )
{
	updateAudioAndTime();
	shader.Use();
	sendDefaultParams( shader );
	sendParams( shader );
	update();
	FFResult result = render( inputTextures );
	consumeAllTrigger();
	resetOpenGLState();

	return result;
}

FFResult Plugin::DeInitGL()
{
	shader.FreeGLResources();
	quad.Release();
	clean();
	return FF_SUCCESS;
}

FFResult Plugin::render( ProcessOpenGLStruct* inputTextures )
{
	shader.Use();
	quad.Draw();
	return FF_SUCCESS;
}

std::string Plugin::createFragmentShader( std::string base )
{
	std::string fragmentShaderCode = fragmentShaderCodeStart;
	int i                          = 0;
	while( i < params.size() )
	{
		Param* param                   = params[ i ];
		const FFParameterType type     = param->getType();
		const std::string& uniformName = param->getUniformName();

		if( isRGBColor( i ) )
		{
			fragmentShaderCode += "uniform vec3 " + uniformName + ";\n";
			i += 3;
		}
		else if( isHueColor( i ) )
		{
			fragmentShaderCode += "uniform vec4 " + uniformName + ";\n";
			i += 4;
		}
		else if( type == FFParameterType::FF_TYPE_BOOLEAN || type == FFParameterType::FF_TYPE_EVENT )
		{
			fragmentShaderCode += "uniform bool " + uniformName + ";\n";
			i++;
		}
		else if( type == FFParameterType::FF_TYPE_INTEGER )
		{
			fragmentShaderCode += "uniform int " + uniformName + ";\n";
			i++;
		}
		else if( type == FFParameterType::FF_TYPE_OPTION )
		{
			// We could handle this in the final "else" case; making explicit case to match with sendParams()
			fragmentShaderCode += "uniform float " + uniformName + ";\n";
			i++;
		}
		else if( type == FFParameterType::FF_TYPE_BUFFER )
		{
			const FFParameterUsage usage = GetParamUsage( param->getIndex() );

			// Check usage
			if( usage == FFParameterUsage::FF_USAGE_FFT )
			{
				// Passed as a vec4 (volume, bass, med, high)
				fragmentShaderCode += "uniform vec4 " + uniformName + ";\n";
			}
			else
			{
				// TODO: FFParameterUsage::FF_USAGE_STANDARD not handled
			}

			i++;
		}
		else if( type == FFParameterType::FF_TYPE_TEXT )
		{
			// Text parameters not passed to shader
			i++;
		}
		else
		{
			// "Normal" values (floats, 0.0 to 1.0, including non-grouped colour parameters), float range values
			fragmentShaderCode += "uniform float " + uniformName + ";\n";
			i++;
		}
	}

	for( auto snippet : includedSnippets )
	{
		fragmentShaderCode += shader::snippets.find( snippet )->second;
	}

	fragmentShaderCode += base;
	return fragmentShaderCode;
}

void Plugin::setParamaterUniformLocations()
{
	for( int i = 0; i < params.size(); i++ )
	{
		Param* param                   = params[ i ];
		const std::string& uniformName = param->getUniformName();
		if( !uniformName.empty() )
		{
			const GLint location = shader.FindUniform( uniformName );

			if(location != -1)
			{
				param->setUniformLocation( location );
			}
			else
			{
#ifdef LOGSHADERNOTES
				printf( "Note: Parameter [%s]: no shader uniform [%s]\n", param->getName().c_str(), uniformName.c_str() );
#endif
			}
		}
	}
}

void Plugin::updateAudioAndTime()
{
	// Update time and frame data
	frame++;
	auto t_now = std::chrono::high_resolution_clock::now();
	timeNow    = std::chrono::duration< float, std::milli >( t_now - t_start ).count() / 1000.0f;
	deltaTime  = timeNow - lastUpdate;
	lastUpdate = timeNow;

	// Update FFT data
	for( auto entry : audioParams )
	{
		ParamFFT* paramFFT = entry.first;
		ParamInfo* fftInfo = FindParamInfo( paramFFT->getIndex() );

		//TODO: use vector copy instead of element by element copy?
		for( size_t index = 0; index < paramFFT->fftData.size(); ++index )
			paramFFT->fftData[ index ] = fftInfo->elements[ index ].value;

		audioParams[ paramFFT ].update( paramFFT->fftData );
	}
}

// 2019-07-12 ffgl@subpixels.com
// - handle FFParameterType::FF_TYPE_BUFFER
// - handle FFParameterType::FF_TYPE_OPTION
void Plugin::sendParams( FFGLShader& shader )
{
	// Clamp to edge is broken in Resolume right now so disable it
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER );
	int i = 0;
	while( i < params.size() )
	{
		Param* param                = params[ i ];
		FFParameterType type        = param->getType();
		const GLint uniformLocation = param->getUniformLocation();

		if( isRGBColor( i ) )
		{
			// Uniform is taken from the first of three parameters
			if( uniformLocation != -1 )
			{
				// RGB only; no alpha
				ffglex::GlColor rgb;

				rgb.r = params[ i ]->getValue();
				rgb.g = params[ i + 1 ]->getValue();
				rgb.b = params[ i + 2 ]->getValue();

				shader.Set( uniformLocation, rgb.r, rgb.g, rgb.b );
			}

			i += 3;
		}
		else if( isHueColor( i ) )
		{
			// Uniform is taken from the first of four parameters
			if( uniformLocation != -1 )
			{
				ffglex::HSBColor hsba;
				hsba.hue   = params[ i ]->getValue();
				hsba.sat   = params[ i + 1 ]->getValue();
				hsba.bri   = params[ i + 2 ]->getValue();
				hsba.alpha = params[ i + 3 ]->getValue();

				//we need to make sure the hue doesn't reach 1.0f, otherwise the result will be pink and not red how it should be
				if( hsba.hue == 1.0f )
					hsba.hue = 0.0f;

				ffglex::GlColor rgba;
				HSBtoRGB( hsba, rgba );

				shader.Set( uniformLocation, rgba.r, rgba.g, rgba.b, rgba.a );
			}

			i += 4;
		}
		else if( type == FFParameterType::FF_TYPE_BOOLEAN || type == FFParameterType::FF_TYPE_EVENT )
		{
			if( uniformLocation != -1 )
			{
				shader.Set( uniformLocation, (bool)param->getValue() );
			}

			i++;
		}
		else if( type == FFParameterType::FF_TYPE_INTEGER )
		{
			// Integers are defined with a range, but the value is not normalised (0.0 to 1.0)
			if( uniformLocation != -1 )
			{
				shader.Set( uniformLocation, (int)param->getValue() );
			}

			i++;
		}
		else if( type == FFParameterType::FF_TYPE_OPTION )
		{
			if( uniformLocation != -1 )
			{
				const auto paramOption = dynamic_cast< ParamOption* >( param );
				assert( paramOption != nullptr );

				shader.Set( uniformLocation, paramOption->getRealValue() );
			}

			i++;
		}
		else if( type == FFParameterType::FF_TYPE_BUFFER )
		{
			if( uniformLocation != -1 )
			{
				const FFParameterUsage usage = GetParamUsage( param->getIndex() );

				// Check usage
				if( usage == FFParameterUsage::FF_USAGE_FFT )
				{
					// Note: Buffer data (elements[]) is copied to fftData[] in updateAudioAndTime()
					const auto paramFFT = dynamic_cast< ParamFFT* >( param );
					assert( paramFFT != nullptr );

					// Locate the related Audio object and extract the volume, bass, med and high values
					Audio audio = audioParams[ paramFFT ];

					// Passed as a vec4 (volume, bass, med, high)
					shader.Set( uniformLocation, audio.getVolume(), audio.getBass(), audio.getMed(), audio.getHigh() );
				}
				else
				{
					// TODO: FFParameterUsage::FF_USAGE_STANDARD not handled
				}
			}

			i++;
		}
		else if( type == FFParameterType::FF_TYPE_TEXT )
		{
			// Text parameters not passed to shader
			i++;
		}
		else
		{
			// "Normal" values (floats, 0.0 to 1.0, including non-grouped colour parameters), float range values
			if( uniformLocation != -1 )
			{
				const auto paramRange = dynamic_cast< ParamRange* >( param );
				float value           = paramRange ? paramRange->getRealValue() : param->getValue();
				shader.Set( uniformLocation, value );
			}

			i++;
		}
	}
}

void Plugin::sendDefaultParams( ffglex::FFGLShader& shader )
{
	shader.Set( "time", timeNow );
	shader.Set( "deltaTime", deltaTime );
	shader.Set( "frame", frame );
	shader.Set( "resolution", (float)currentViewport.width, (float)currentViewport.height );
	shader.Set( "bpm", bpm );
	shader.Set( "phase", barPhase );
}

char* Plugin::GetParameterDisplay( unsigned int index )
{
	bool inRange = 0 <= index && index < params.size();

	// 2019-08-01 ffgl@subpixels.com - valid was testing against FF_TYPE_TEXT twice; changed to check for text and buffer types
	Param* param              = params[ index ];
	FFParameterType paramType = param->getType();
	bool valid                = paramType != FFParameterType::FF_TYPE_TEXT && paramType != FFParameterType::FF_TYPE_BUFFER;

	if( inRange && valid )
	{
		// 2019-07-10 ffgl@subpixels.com - be careful not to copy over the last character (the null-terminator) in displayValueBuffer
		static char displayValueBuffer[ 16 ];
		auto range              = dynamic_cast< ParamRange* >( param );
		float value             = range ? range->getRealValue() : params[ index ]->getValue();
		std::string stringValue = std::to_string( value );
		memset( displayValueBuffer, 0, sizeof( displayValueBuffer ) );
		memcpy( displayValueBuffer, stringValue.c_str(), std::min( sizeof( displayValueBuffer ) - 1, stringValue.length() ) );
		return displayValueBuffer;
	}
	else
	{
		return (char*)FF_FAIL;
	}
}

FFResult Plugin::SetFloatParameter( unsigned int index, float value )
{
	if( index < params.size() )
	{
		params[ index ]->setValue( value );
		return FF_SUCCESS;
	}
	else
	{
		return FF_FAIL;
	}
}

float Plugin::GetFloatParameter( unsigned int index )
{
	if( 0 <= index && index < params.size() )
	{
		return params[ index ]->getValue();
	}
	else
	{
		return 0.0f;
	}
}

// 2019-07-17 ffgl@subpixels.com - use setText()
FFResult Plugin::SetTextParameter( unsigned int index, const char* value )
{
	bool inBounds = 0 <= index && index < params.size();
	if( !inBounds )
		return FF_FAIL;

	auto paramText = dynamic_cast< ParamText* >( params[ index ] );
	if( !paramText )
		return FF_FAIL;

	paramText->setText( value == nullptr ? "" : value );
	return FF_SUCCESS;
}

// 2019-07-17 ffgl@subpixels.com - use getText()
char* Plugin::GetTextParameter( unsigned int index )
{
	bool inBounds = 0 <= index && index < params.size();
	if( !inBounds )
		return "";

	auto paramText = dynamic_cast< ParamText* >( params[ index ] );
	if( !paramText )
		return "";

	// Use this static to keep the returned string around for the caller to access the char data
	static std::string returnStr;
	returnStr = paramText->getText();

	return const_cast< char* >( returnStr.c_str() );
}

void Plugin::SetSampleRate( unsigned int _sampleRate )
{
	sampleRate = _sampleRate;
	for( auto entry : audioParams )
	{
		ParamFFT* param = entry.first;
		audioParams[ param ].setSampleRate( _sampleRate );
	}
}

void Plugin::setFragmentShader( std::string base )
{
	fragmentShaderBase = base;
}

void Plugin::addParam( Param* param )
{
	unsigned int newIndex = (unsigned int)params.size();
	param->setIndex( newIndex );
	SetParamInfo( newIndex, param->getName().c_str(), param->getType(), param->getValue() );
	params.push_back( param );
}

void Plugin::addParam( ParamRange* param )
{
	unsigned int newIndex = (unsigned int)params.size();
	param->setIndex( newIndex );
	SetParamInfo( newIndex, param->getName().c_str(), param->getType(), param->getValue() );
	SetParamRange( newIndex, param->getRange().min, param->getRange().max );
	params.push_back( param );
}

void Plugin::addParam( ParamOption* param )
{
	unsigned int newIndex = (unsigned int)params.size();
	param->setIndex( newIndex );
	SetOptionParamInfo( newIndex, param->getName().c_str(), (unsigned int)param->options.size(), param->getValue() );

	for( unsigned int i = 0; i < param->options.size(); i++ )
	{
		//TODO: the elementValue argument ((float)i) makes options numbered from 0; why are we ignoring options[i].value ?
		SetParamElementInfo( newIndex, i, param->options[ i ].name.c_str(), (float)i );
	}
	params.push_back( param );
}

void Plugin::addParam( ParamFFT* param )
{
	unsigned int newIndex = (unsigned int)params.size();
	param->setIndex( newIndex );
	audioParams[ param ] = Audio();
	SetBufferParamInfo( newIndex, param->getName().c_str(), (unsigned int)param->fftData.size(), FFParameterUsage::FF_USAGE_FFT );
	params.push_back( param );
}


// 2019-08-13 ffgl@subpixels.com - Body extracted to addHueColorParam(name, uniformName) constructor
void Plugin::addHueColorParam( std::string name )
{
	addHueColorParam( name, name );
}
void Plugin::addHueColorParam( std::string name, std::string uniformName )
{
	addParam( new Param( name, FFParameterType::FF_TYPE_HUE, 0.0f, uniformName ) );
	addParam( new Param( name + " S", FFParameterType::FF_TYPE_SATURATION, 1.0f, std::string() ) );
	addParam( new Param( name + " B", FFParameterType::FF_TYPE_BRIGHTNESS, 1.0f, std::string() ) );
	addParam( new Param( name + " A", FFParameterType::FF_TYPE_ALPHA, 1.0f, std::string() ) );
}

// 2019-08-13 ffgl@subpixels.com - Body extracted to addRGBColorParam(name, uniformName) constructor
void Plugin::addRGBColorParam( std::string name )
{
	addRGBColorParam(name, name);
}
// 2019-08-13 ffgl@subpixels.com
// - Change plugin paramater titles from ("name", "name_g", "name_b") to ("name R", "name G", "name B")
// - Use Param( name, type, value, uniformName ) constructor, no uniform names for the G and B parameters
void Plugin::addRGBColorParam( std::string name, std::string uniformName )
{
	addParam( new Param( name + " R", FFParameterType::FF_TYPE_RED, 0.5f, uniformName ) );
	addParam( new Param( name + " G", FFParameterType::FF_TYPE_GREEN, 0.5f, std::string() ) );
	addParam( new Param( name + " B", FFParameterType::FF_TYPE_BLUE, 0.5f, std::string() ) );
}

bool Plugin::isHueColor( int index )
{
	bool enoughSpace = index + 3 < params.size();
	if( !enoughSpace )
		return false;
	bool isColorType =
		params[ index ]->getType() == FFParameterType::FF_TYPE_HUE &&
		params[ index + 1 ]->getType() == FFParameterType::FF_TYPE_SATURATION &&
		params[ index + 2 ]->getType() == FFParameterType::FF_TYPE_BRIGHTNESS &&
		params[ index + 3 ]->getType() == FFParameterType::FF_TYPE_ALPHA;

	return isColorType;
}

bool Plugin::isRGBColor( int index )
{
	bool enoughSpace = index + 2 < params.size();
	if( !enoughSpace )
		return false;
	bool isColorType =
		params[ index ]->getType() == FFParameterType::FF_TYPE_RED &&
		params[ index + 1 ]->getType() == FFParameterType::FF_TYPE_GREEN &&
		params[ index + 2 ]->getType() == FFParameterType::FF_TYPE_BLUE;

	return isColorType;
}

Param* Plugin::getParam( std::string name )
{
	for( int i = 0; i < params.size(); i++ )
	{
		if( params[ i ]->getName().compare( name ) == 0 )
			return params[ i ];
	}
	return {};
}

ParamOption* Plugin::getParamOption( std::string name )
{
	auto param = getParam( name );
	if( !param )
		return {};
	auto option = dynamic_cast< ParamOption* >( param );
	if( !option )
		return {};
	return option;
}

ParamText* Plugin::getParamText( std::string name )
{
	auto param = getParam( name );
	if( !param )
		return {};
	auto text = dynamic_cast< ParamText* >( param );
	if( !text )
		return {};
	return text;
}

void Plugin::include( shader::snippet_id snippet )
{
	if( includedSnippets.find( snippet ) != includedSnippets.end() )
		return;

	includedSnippets.insert( snippet );
	auto deps = shader::dependencies.find( snippet );
	if( deps == shader::dependencies.end() )
		return;
	for( auto dep : deps->second )
	{
		includedSnippets.insert( dep );
	}
}

void Plugin::include( std::set< shader::snippet_id > snippets )
{
	for( auto snippet : snippets )
	{
		include( snippet );
	}
}

void Plugin::consumeAllTrigger()
{
	for( int i = 0; i < params.size(); i++ )
	{
		auto trigger = dynamic_cast< ParamTrigger* >( params[ i ] );
		if( trigger )
		{
			trigger->consume();
		}
	}
}
